#include <stdio.h>
#include <zlib.h>
#include "../bzlib.h"
#include <memory>
#include <vector>


typedef std::vector<unsigned char> UCVECTOR;

class GYM
{
public:
	enum {
		TYPE_LAW,
		TYPE_WITHHEADER,
		TYPE_DEFLATED,
		TYPE_EZPACKED
	} type;
	enum {
		OFS_MAGIC = 0,
		OFS_SONGTITLE = (OFS_MAGIC +      4),
		OFS_GAMETITLE = (OFS_SONGTITLE + 32),
		OFS_PUBLISHER = (OFS_GAMETITLE + 32),
		OFS_GENERATOR = (OFS_PUBLISHER + 32),
		OFS_DUMPER    = (OFS_GENERATOR + 32),
		OFS_COMMENTS  = (OFS_DUMPER +    32),
		OFS_LOOPTIME  = (OFS_COMMENTS + 256),
		OFS_COMPSIZE  = (OFS_LOOPTIME +   4),
		OFS_DATASTART = (OFS_COMPSIZE +   4)
	};
	enum
	{
		MAGIC_GYMX = (0x584d5947UL),	/* 'GYMX' */
		MAGIC_EZPK = (0x4b505a45UL)		/* 'EZPK' */
	};

	unsigned char header[OFS_DATASTART];
	UCVECTOR data;

	GYM()
	{
		memset(header, 0, sizeof(header));
	};
	~GYM()
	{
	};
};

class FILEREADER
{
protected:
	FILE *m_fp;
	unsigned long m_length;
public:
	FILEREADER(const char *path)
	{
		m_fp = fopen(path, "rb");
		if (m_fp)
		{
			fseek(m_fp, 0, SEEK_END);
			m_length = ftell(m_fp);
			fseek(m_fp, 0, SEEK_SET);
		}
		else
			m_length = 0;
	}
	~FILEREADER()
	{
		if (m_fp) fclose(m_fp);
	}
	bool ready()
	{
		return m_fp != 0;
	}
	unsigned long length()
	{
		return m_length;
	}
	unsigned long read(unsigned char *buf, unsigned len)
	{
		return m_fp ? fread(buf, 1, len, m_fp) : 0;
	}
};

class FILEWRITER
{
protected:
	FILE *m_fp;
public:
	FILEWRITER(const char *path)
	{
		m_fp = fopen(path, "wb");
	}
	~FILEWRITER()
	{
		if (m_fp) fclose(m_fp);
	}
	unsigned long write(unsigned char *buf, unsigned len)
	{
		return m_fp ? fwrite(buf, 1, len, m_fp) : 0;
	}
};

static unsigned long GetDwordLE(const unsigned char *p)
{
	unsigned long ret;
	ret  = ((unsigned long)(const unsigned char)p[3]) << 0x18;
	ret |= ((unsigned long)(const unsigned char)p[2]) << 0x10;
	ret |= ((unsigned long)(const unsigned char)p[1]) << 0x08;
	ret |= ((unsigned long)(const unsigned char)p[0]) << 0x00;
	return ret;
}

static void SetDwordLE(unsigned char *p, unsigned long v)
{
	p[0] = (unsigned char)(v >> (8 * 0));
	p[1] = (unsigned char)(v >> (8 * 1));
	p[2] = (unsigned char)(v >> (8 * 2));
	p[3] = (unsigned char)(v >> (8 * 3));
}

static unsigned long decomp_dac(unsigned char *srcp, unsigned long srcl, unsigned char *desp, unsigned long desl)
{
	unsigned long ul;
	if (GetDwordLE(srcp + 0) != GYM::MAGIC_EZPK) return 0;
	ul = GetDwordLE(srcp + 4);
	if (desl == 0) return ul & 0xfffffff;
	if ((ul & 0xf0000000) == 0)
	{
		unsigned int ui = desl;
		if (BZ2_bzBuffToBuffDecompress((char *)desp, &ui, (char *)(srcp + 8), srcl, 0, 0) < 0)
			return 0;
		return desl;
	}
	else
	{
		z_stream zs;
		memset(&zs, 0, sizeof(z_stream));
		zs.next_in = (Bytef *)(srcp + 8);
		zs.avail_in = srcl - 8;
		zs.next_out = (Bytef *)desp;
		zs.avail_out = desl;
		if ((inflateInit(&zs)) < 0) return 0;
		if ((inflate(&zs, Z_FINISH)) >= 0)
			ul = zs.total_out;
		else
			ul = 0;
		inflateEnd(&zs);
		return ul;
	}
}


GYM * load_from_file(const char *path)
{
	FILEREADER fr(path);
	if (!fr.ready()) return 0;
	GYM *gym = new GYM;
	if (!gym) return 0;
	unsigned long ul = fr.read(gym->header, GYM::OFS_DATASTART);
	if (ul < GYM::OFS_DATASTART || GetDwordLE(gym->header) != GYM::MAGIC_GYMX)
	{
		/* wb_ */
		gym->type = GYM::TYPE_LAW;
		gym->data.resize(fr.length());
		memcpy(&gym->data[0], gym->header, ul);
		if (ul == GYM::OFS_DATASTART)
			fr.read(&gym->data[ul], fr.length() - ul);
		memset(gym->header, 0, sizeof(gym->header));
		SetDwordLE(gym->header, GYM::MAGIC_GYMX);
		return gym;
	}
	unsigned long xsize = GetDwordLE(gym->header + GYM::OFS_COMPSIZE);
	if (xsize == 0)
	{
		/* k */
		gym->type = GYM::TYPE_WITHHEADER;
		gym->data.resize(fr.length() - ul);
		fr.read(&gym->data[0], fr.length() - ul);
		return gym;
	}
	gym->data.resize(xsize);

	SetDwordLE(gym->header + GYM::OFS_COMPSIZE, 0);

	UCVECTOR cmpdbuf_v;
	cmpdbuf_v.resize(fr.length() - ul);
	fr.read(&cmpdbuf_v[0], fr.length() - ul);
	if (GetDwordLE(&cmpdbuf_v[0]) != GYM::MAGIC_EZPK)
	{
		/* inflate */
		uLongf dessize;
		gym->type = GYM::TYPE_DEFLATED;
		uncompress((Bytef *)(&gym->data[0]), (uLongf *)(&dessize), (const Bytef *)(&cmpdbuf_v[0]), (uLong)(fr.length() - ul));
		return gym;
	}

	/* EZPK */
	gym->type = GYM::TYPE_EZPACKED;

	UCVECTOR unezbuf_v;
	unsigned int unezbuf_l = xsize + 0x24;
	unezbuf_v.resize(xsize + 0x24);
	BZ2_bzBuffToBuffDecompress((char *)(&unezbuf_v[0]), &unezbuf_l, (char *)(&cmpdbuf_v[8]), GetDwordLE(&cmpdbuf_v[4]) - 8, 0, 0);
	unsigned long undacbuf_l = GetDwordLE(&unezbuf_v[0x18]) + GetDwordLE(&unezbuf_v[0x1c]);
	if (undacbuf_l >= 100)
	{
		UCVECTOR undacbuf_v;
		undacbuf_v.resize(undacbuf_l);
		unsigned long ret = decomp_dac(&unezbuf_v[GetDwordLE(&unezbuf_v[0x20])],GetDwordLE(&unezbuf_v[0x14]), &undacbuf_v[0], undacbuf_l);
		if (ret) memcpy(&unezbuf_v[GetDwordLE(&unezbuf_v[0x20])], &undacbuf_v[0], ret);
	}

	unsigned char *typep, *sngp, *regp, *datmp, *datsp, *dacmp, *dacsp;
	{
		unsigned char *p1;
		p1 = &unezbuf_v[0x24];
		typep = p1; p1 += GetDwordLE(&unezbuf_v[0x00]);
		regp  = p1; p1 += GetDwordLE(&unezbuf_v[0x04]);
		datmp = p1; p1 += GetDwordLE(&unezbuf_v[0x08]);
		datsp = p1; p1 += GetDwordLE(&unezbuf_v[0x0c]);
		sngp  = p1; p1 += GetDwordLE(&unezbuf_v[0x10]);
		dacmp = p1; p1 += GetDwordLE(&unezbuf_v[0x18]);
		dacsp = p1;
	}
	unsigned idx = 0;
	do {
		unsigned char uc;
		gym->data[idx++] = uc = *(typep++);
		switch (uc)
		{
			case 1:
				gym->data[idx++] = uc = *(regp++);
				if (uc == 0x2a)
					gym->data[idx++] = *(dacmp++);
				else
					gym->data[idx++] = *(datmp++);
				break;
			case 2:
				gym->data[idx++] = uc = *(regp++);
				if (uc == 0x2a)
					gym->data[idx++] = *(dacsp++);
				else
					gym->data[idx++] = *(datsp++);
				break;
			case 3:
				gym->data[idx++] = *(sngp++);
				break;
		}
	} while (idx < xsize);
	return gym;
}

class EZPK
{
public:
	unsigned char header[GYM::OFS_DATASTART];
	UCVECTOR pkbuf_v;
	void ezpk(GYM *gym, int *lvl)
	{
		enum {
			PHASE_TYPE = 0,
			PHASE_REG,
			PHASE_DATM,
			PHASE_DATS,
			PHASE_SNG,
			PHASE_DACM,
			PHASE_DACS,
			PHASE_NUM
		};
		memcpy(header, gym->header, GYM::OFS_DATASTART);
		unsigned long size[PHASE_NUM], base[PHASE_NUM];
		unsigned long ofs = 0x24, datsize = gym->data.size();
		/* 32oCg߂Ɋm */
		UCVECTOR sortbuf_v;
		sortbuf_v.resize(0x24 + datsize + 32);
		for (int phase = 0; phase < PHASE_NUM; phase++)
		{
			const unsigned long iend = datsize;
			unsigned long icur = 0;
			base[phase] = ofs;
			do
			{
				unsigned char uc;
				uc = gym->data[icur++];
				if (phase == PHASE_TYPE) sortbuf_v[ofs++] = uc;
				switch (uc)
				{
				case 1:
					if (icur == iend) break;
					uc = gym->data[icur++];
					if (phase == PHASE_REG) sortbuf_v[ofs++] = uc;
					if (icur == iend) break;
					if (uc == 0x2a)
					{
						uc = gym->data[icur++];
						if (phase == PHASE_DACM) sortbuf_v[ofs++] = uc;
					}
					else
					{
						uc = gym->data[icur++];
						if (phase == PHASE_DATM) sortbuf_v[ofs++] = uc;
					}
					break;
				case 2:
					if (icur == iend) break;
					uc = gym->data[icur++];
					if (phase == PHASE_REG) sortbuf_v[ofs++] = uc;
					if (icur == iend) break;
					if (uc == 0x2a)
					{
						uc = gym->data[icur++];
						if (phase == PHASE_DACS) sortbuf_v[ofs++] = uc;
					}
					else
					{
						uc = gym->data[icur++];
						if (phase == PHASE_DATS) sortbuf_v[ofs++] = uc;
					}
					break;
				case 3:
					if (icur == iend) break;
					uc = gym->data[icur++];
					if (phase == PHASE_SNG) sortbuf_v[ofs++] = uc;
					break;
				}
			} while (icur < iend);
			size[phase] = ofs - base[phase];
		}
		SetDwordLE(&sortbuf_v[0x00], size[PHASE_TYPE]);
		SetDwordLE(&sortbuf_v[0x04], size[PHASE_REG]);
		SetDwordLE(&sortbuf_v[0x08], size[PHASE_DATM]);
		SetDwordLE(&sortbuf_v[0x0c], size[PHASE_DATS]);
		SetDwordLE(&sortbuf_v[0x10], size[PHASE_SNG]);
		SetDwordLE(&sortbuf_v[0x14], 0);
		SetDwordLE(&sortbuf_v[0x18], size[PHASE_DACM]);
		SetDwordLE(&sortbuf_v[0x1c], size[PHASE_DACS]);
		SetDwordLE(&sortbuf_v[0x20], 0);
		/* TCYĒ */
		sortbuf_v.resize(ofs);
		unsigned long dacbase = base[PHASE_DACM];
		unsigned long dacsize = size[PHASE_DACM] + size[PHASE_DACS];
		if ((lvl[0] || lvl[2]) && dacsize >= 100)
		{
			if (lvl[2])
			{
				unsigned int dacout = (dacsize << 1) + 12;
				UCVECTOR dacbuf_v;
				dacbuf_v.resize(dacout);
				if (BZ_OK == BZ2_bzBuffToBuffCompress((char *)(&dacbuf_v[0]), &dacout, (char *)(&sortbuf_v[dacbase]), dacsize, lvl[2], 0, 0) && dacout + 8 < dacsize)
				{
					SetDwordLE(&sortbuf_v[0x14], dacout + 8);
					SetDwordLE(&sortbuf_v[0x20], dacbase);
					SetDwordLE(&sortbuf_v[dacbase + 0], GYM::MAGIC_EZPK);
					SetDwordLE(&sortbuf_v[dacbase + 4], (dacout + 8) | 0x00000000);
					memcpy(&sortbuf_v[dacbase + 8], &dacbuf_v[0], dacout);
					/* TCYĒ */
					sortbuf_v.resize(dacbase + 8 + dacout);
					dacsize = 0;
				}
			}
			else
			{
				uLong dacout = (dacsize << 1) + 12;
				UCVECTOR dacbuf_v;
				dacbuf_v.resize(dacout);
				if (Z_OK == compress2((Bytef *)(&dacbuf_v[0]), &dacout, (const Bytef *)(&sortbuf_v[dacbase]), dacsize, lvl[0]) && dacout + 8 < dacsize)
				{
					SetDwordLE(&sortbuf_v[0x14], dacout + 8);
					SetDwordLE(&sortbuf_v[0x20], dacbase);
					SetDwordLE(&sortbuf_v[dacbase + 0], GYM::MAGIC_EZPK);
					SetDwordLE(&sortbuf_v[dacbase + 4], (dacout + 8) | 0x10000000);
					memcpy(&sortbuf_v[dacbase + 8], &dacbuf_v[0], dacout);
					/* TCYĒ */
					sortbuf_v.resize(dacbase + 8 + dacout);
					dacsize = 0;
				}
			}
		}
		if (dacsize >= 100)
		{
			/* k */
			SetDwordLE(&sortbuf_v[0x1c], (-size[PHASE_DACM]));
		}
		UCVECTOR bzbuf_v;
		unsigned int bzbuf_l = sortbuf_v.size() << 1;
		bzbuf_v.resize(bzbuf_l);
		if (BZ_OK == BZ2_bzBuffToBuffCompress((char *)(&bzbuf_v[0]), &bzbuf_l, (char *)(&sortbuf_v[0]), sortbuf_v.size(), lvl[1], 0, 0) && bzbuf_l + 8 < datsize)
		{
			pkbuf_v.resize(bzbuf_l + 8);
			SetDwordLE(&pkbuf_v[0], GYM::MAGIC_EZPK);
			SetDwordLE(&pkbuf_v[4], bzbuf_l + 8);
			memcpy(&pkbuf_v[8], &bzbuf_v[0], bzbuf_l);
			SetDwordLE(&header[GYM::OFS_COMPSIZE], datsize);
		}
		else
		{
			/* ǖk */
			pkbuf_v.resize(datsize);
			memcpy(&pkbuf_v[0], &gym->data[0], datsize);
			SetDwordLE(&header[GYM::OFS_COMPSIZE], 0);
		}
	}
};


void save_to_file_withoutheader(const char *path, GYM *gym)
{
	FILEWRITER fw(path);
	fw.write(&gym->data[0], gym->data.size());
}

void save_to_file_withheader(const char *path, GYM *gym)
{
	FILEWRITER fw(path);
	fw.write(gym->header, GYM::OFS_DATASTART);
	fw.write(&gym->data[0], gym->data.size());
}

void save_to_file_ezpk(const char *path, GYM *gym, int *lvl)
{
	EZPK ezpk;
	ezpk.ezpk(gym, lvl);

#if 1
	FILEWRITER fw(path);
#else
	FILEWRITER fw("debug.gym");
#endif
	fw.write(ezpk.header, GYM::OFS_DATASTART);
	fw.write(&ezpk.pkbuf_v[0], ezpk.pkbuf_v.size());
}

void gymcmp(const char *path, bool decompmode, int *lvl)
{
	std::auto_ptr<GYM> gym(load_from_file(path));
	if (gym.get() == 0)
	{
			fprintf(stderr, "ǂݍݎs : [ %s ]\n", path);
			return;
	}
	if (decompmode)
	{
		if (gym->type <= GYM::TYPE_WITHHEADER)
		{
			fprintf(stderr, "WJς : [ %s ] %s\n", path, gym->header + GYM::OFS_SONGTITLE);
			return;
		}
		save_to_file_withheader(path, gym.get());
		fprintf(stderr, "WJ : [ %s ] %s\n", path, gym->header + GYM::OFS_SONGTITLE);
		return;
	}
#if 0
	if (gym->type == GYM::TYPE_EZPACKED)
	{
		fprintf(stderr, "EZPKkς : [ %s ] %s\n", path, gym->header + GYM::OFS_SONGTITLE);
		return;
	}
#endif
	save_to_file_ezpk(path, gym.get(), lvl);
	fprintf(stderr, "EZPKk : [ %s ] %s\n", path, gym->header + GYM::OFS_SONGTITLE);
}

int main(int argc, char **argv)
{
	int i;
	int bz = 1, lvl[3] = { 9, 2, 0 } ;
	bool d = false;
	bool opt = true;
	for (i = 1; i < argc; i++)
	{
		if (opt && argv[i][0] == '-')
		{
			for (int j = 1; argv[i][j]; j++)
			{
				switch (argv[i][j])
				{
				case '0':	case '1':	case '2':	case '3':	case '4':
				case '5':	case '6':	case '7':	case '8':	case '9':
					lvl[bz] = argv[i][j] - '0';
					bz = 1;
					break;
				case 'E':	case 'e':
					bz = 1;
					break;
				case 'Z':	case 'z':
					bz = 0;
					break;
				case 'B':	case 'b':
					/* \ */
					bz = 2;
					break;
				case 'D':	case 'd':
					d = true;
					break;
				case 'C':	case 'c':
					d = false;
					break;
				case '-':
					opt = false;
					break;
				}
			}
		}
		else
		{
			gymcmp(argv[i], d, lvl);
		}
	}
	return 0;
}
